use std::{io::{BufRead, stdin}, fmt::Display};

use myutils::matrix::Matrix;
use thiserror::Error;
use anyhow::{Result, anyhow};

#[derive(Debug, Error)]
#[error("Invalid tile {0}")]
struct InvalidTile(char);

#[derive(Debug, Clone, Copy, Eq, PartialEq)]
enum Direction {
    N,E,S,W
}

#[derive(Debug,Clone,Copy)]
struct Tile { n: bool, e: bool, s: bool, w: bool }

const EMPTY: Tile = Tile { n: false, e: false, s: false, w: false };

impl Tile {

    fn connects(&self, dir: Direction) -> bool {
        use Direction::*;
        match dir {
            N => self.n,
            E => self.e,
            S => self.s,
            W => self.w,
        }
    }

    fn follow(&self, dir: Direction) -> Option<Direction> {
        use Direction::*;
        //if !self.connects(dir) { return None }
        for &exit in &[N,E,S,W] {
            if exit != dir && self.connects(exit) {
                return Some(exit)
            }
        }
        None
    }
}

impl TryFrom<char> for Tile {
    type Error = InvalidTile;

    fn try_from(value: char) -> Result<Self, Self::Error> {
        let (n,e,s,w) = match value {
            '|' => (true, false, true, false),
            '-' => (false, true, false, true),
            'L' => (true, true, false, false),
            'J' => (true, false, false, true),
            '7' => (false, false, true, true),
            'F' => (false, true, true, false),
            '.' => (false, false, false, false),
            other => return Err(InvalidTile(other)),
        };
        Ok(Tile{n,e,s,w})
    }
}

impl Tile {
    fn char(&self) -> char {
        match (self.n, self.e, self.s, self.w) {
            (false, true, false, true) => '-',
            (true, false, true, false) => '|',
            (true, true, false, false) => 'L',
            (true, false, false, true) => 'J',
            (false, false, true, true) => '7',
            (false, true, true, false) => 'F',
            (false, false, false, false) => '.',
           _ => '?',
       }     
    }
}

impl Display for Tile {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.char())
    }
}

#[derive(Debug)]
struct Map {
    grid: Matrix<Tile>,
    start: (usize, usize),
}

impl Map {
    fn load(r: &mut dyn BufRead) -> Result<Self> {
        let mut buffer = vec![];
        let mut lines = 0;
        let mut start = None;
        for (x, line) in r.lines().enumerate() {
            let line = line?;
            lines += 1;
            for (y, c) in line.chars().enumerate() {
                if c == 'S' {
                    start = Some((x,y));
                    buffer.push(EMPTY);
                } else {
                    buffer.push(Tile::try_from(c)?)
                }
            }
        }

        let Some(start) = start else { return Err(anyhow!("no start point")) };
        let shape = (lines, buffer.len() / lines);
        let mut grid = Matrix::new(buffer, shape)?;

        let start_tile = Tile {
            n: start.0 > 0 && grid[start.0 - 1][start.1].s,
            e: start.1 + 1 < grid.shape().1 && grid[start.0][start.1 + 1].w,
            s: start.0 + 1 < grid.shape().0 && grid[start.0 + 1][start.1].n,
            w: start.1 > 0 && grid[start.0][start.1 - 1].e,
        };

        grid[start.0][start.1] = start_tile;

        Ok(Self { grid, start })

    }

    fn extract_main_loop(&self) -> Self {

        let mut grid = self.grid.map(|_| EMPTY);

        grid[self.start.0][self.start.1] = self.grid[self.start.0][self.start.1];

        let (s,_) = self.snakes();
        for (x,y) in s.take_while(|p| p != &self.start) {
            grid[x][y] = self.grid[x][y];
        }

        Self { grid, start: self.start }
    }

    fn fill(&self) -> Matrix<bool> {
        let shape = self.grid.shape();
        let mut map = Matrix::constant(false, (shape.0 + 1, shape.1 + 1));

        for x in 1..shape.0 {
            for y in 1..shape.1 {
                let cell = &self.grid[x-1][y-1];
                map[x][y] = map[x-1][y-1] ^ cell.n ^ cell.e;
            }
        }

        map
    }

    

}

impl Display for Map {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let (h,w) = self.grid.shape();
        for x in 0..h {
            for y in 0..w {
                if (x,y) == self.start {
                    write!(f, "S")?;
                } else {
                    write!(f, "{}", self.grid[x][y])?;
                }
            }
            writeln!(f)?;
        }
        Ok(())
    }
}

#[derive(Debug)]
struct Snake<'m> {
    map: &'m Map,
    pos: (usize, usize),
    dir: Direction,
}

impl Map {
    fn snakes(&self) -> (Snake, Snake) {
        let start = self.grid[self.start.0][self.start.1];
        let d1 = start.follow(Direction::W).unwrap();
        let d2 = start.follow(d1).unwrap();
        let s1 = Snake {
            map: self,
            pos: self.start,
            dir: d1
        };
        let s2 = Snake {
            map: self,
            pos: self.start,
            dir: d2,
        };
        (s1, s2)
    }
}

impl Iterator for Snake<'_> {
    type Item = (usize, usize);

    fn next(&mut self) -> Option<Self::Item> {
        use Direction::*;
        let from;
        (from, self.pos) = match self.dir {
            N => (S, (self.pos.0 - 1, self.pos.1)),
            E => (W, (self.pos.0, self.pos.1 + 1)),
            S => (N, (self.pos.0 + 1, self.pos.1)),
            W => (E, (self.pos.0, self.pos.1 - 1)),
        };
        self.dir = self.map.grid[self.pos.0][self.pos.1].follow(from)?;
        //dbg!(&self);
        Some(self.pos)
    }
}

fn main() -> Result<()> {
    let map = Map::load(&mut stdin().lock())?.extract_main_loop();
    eprintln!("{}", map);
    
    let (s1, s2) = map.snakes();

    let len = s1
        .zip(s2)
        .enumerate()
        .find_map(|(c, (a,b))| Some(c+1).filter(|_| a == b)).unwrap();

    println!("Collide at {len}");

    let fillmap = map.fill();

    let mut filled = Matrix::constant('?', map.grid.shape());
    for (x,y) in filled.indices() {
        let p: Tile = map.grid[x][y];
        let f = fillmap[x][y];

        filled[x][y] = if !(p.n || p.e || p.s || p.w) {
            if f { 'I' } else { 'O' }
        } else {
            p.char()
        }
    }

    println!("Filled:\n{}", filled);

    let surface = filled.elems()
        .copied()
        .filter(|&e| e == 'I')
        .count();

    println!("Inner surface: {surface}");

    Ok(())
}
