use std::{str::FromStr, ops::RangeFrom, io::stdin, collections::HashMap, fmt::Display};
use anyhow::{Result, Error, anyhow};

#[derive(Debug,Clone,PartialEq,Eq)]
struct Row {
    bits: Vec<Option<bool>>,
    pattern: Vec<usize>,
    freedom: usize,
}

#[derive(Debug)]
struct RowView<'r> {
    bits: &'r [Option<bool>],
    pattern: &'r [usize],
    freedom: usize,
}

type Cache = HashMap<(usize,usize), usize>;

impl Row {
    fn view(&self) -> RowView<'_> {
        RowView { bits: &self.bits, pattern: &self.pattern, freedom: self.freedom }   
    }

    fn enumerate<F>(&self, f: &mut F)
        where F: FnMut(&[bool])
    {
        let mut buf = vec![false; self.bits.len()];
        self.view().enumerate(&mut buf, 0.., f);
    }

    fn optimize(&mut self) {
        let mut offset = 0;
        for &block in &self.pattern {
            if block > self.freedom {
                for cell in &mut self.bits[offset + self.freedom .. offset + block] {
                    *cell = Some(true)
                }
            }
            offset += block + 1;
        }
    }

    fn multiply(self, n: usize) -> Self {
        let mut bits = self.bits;
        bits.push(None);
        bits = bits.repeat(n);
        bits.pop();

        let pattern = self.pattern.repeat(n);

        Row { bits, pattern, freedom: self.freedom * n}
    }

    fn count(&self) -> usize {
        self.view().count()
    }

    fn exhaustive_count(&self) -> usize {
        self.view().exhaustive_count()
    }

}

impl FromStr for Row {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let (bits, nums) = s.split_once(' ')
            .ok_or_else(|| anyhow!("Bad line {s}"))?;

        let bits = bits.chars()
            .map(|c| match c { 
                '.' => Ok(Some(false)),
                '#' => Ok(Some(true)),
                '?' => Ok(None),
                oth => Err(anyhow!("Invalid character {oth}")),
             })
             .collect::<Result<Vec<_>>>()?;        

        let pattern = nums.split(',')
             .map(str::parse)
             .collect::<Result<Vec<_>, _>>()?;

        let size = (pattern.iter().sum::<usize>() + pattern.len())
             .saturating_sub(1);

        let freedom = bits.len()
             .checked_sub(size).ok_or_else(|| anyhow!("Unsolvable row {s}"))?;

        Ok(Self { bits, pattern, freedom })
    }

}

impl Display for Row {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for bit in &self.bits {
            write!(f, "{}", match bit {
                None => '?',
                Some(false) => '.',
                Some(true) => '#',
            })?;
        }
        write!(f, "{:?}", &self.pattern)?;
        Ok(())
    }
}

impl <'r> RowView<'r> {

    fn exhaustive_count(&self) -> usize {
        let mut c = 0;
        let mut buf = vec![false; self.bits.len()];
        self.enumerate(&mut buf, 0.., &mut |_| c += 1);
        c
    }

    fn count(&self) -> usize {
        self.dfs(0, 0, self.freedom, &mut HashMap::new())
    }

    fn dfs(&self, position: usize, consumed: usize, freedom: usize, cache: &mut Cache) -> usize {
        //eprintln!("DFS enter position={position} consumed={consumed} freedom={freedom}");
        fn cached_dfs(this: &RowView, position: usize, consumed: usize, freedom: usize, cache: &mut Cache) -> usize {
            if let Some(&solutions) = cache.get(&(consumed, position)) {
                //eprintln!("Cache hit for position={position} consumed={consumed} freedom={freedom} = {solutions}");
                return solutions
            }

            let solutions = this.dfs(position, consumed, freedom, cache);
            cache.insert((consumed, position), solutions);
            solutions
        }

        if consumed == self.pattern.len() {
                if position <= self.bits.len() && self.bits[position..].iter().any(|b| b == &Some(true)) {
                    //eprintln!("Terminal conflict");
                    0 
                } else { 
                    //eprintln!("Valid solution");
                    1 
                }
        } else {
            let current = self.pattern[consumed];
            let mut solutions = 0;
            for offset in 0..=freedom {
                // Uncovered bit in prefix space, no point in continuing
                if offset > 0 && self.bits[position+offset-1] == Some(true) { break }

                // Block collides with empty cell, try next offset
                if self.bits[position+offset..][..current].iter().any(|b| b == &Some(false)) { continue }

                // Block is not last and cell following block is not empty
                if position+offset+current < self.bits.len() && self.bits[position+offset+current] == Some(true) { continue }
                
                solutions += cached_dfs(&self, position+offset+current+1, consumed+1, freedom - offset, cache);
            }
            //eprintln!("DFS leave position={position} consumed={consumed} freedom={freedom} -> {solutions}");
            solutions
        }

    }

    fn enumerate<F>(&self, buf: &mut [bool], window: RangeFrom<usize>, f: &mut F)
        where F: FnMut(&[bool])
    {
        if let Some((&h, rest)) = self.pattern.split_first() {
            for offset in 0..=self.freedom {

                if self.bits[..offset].iter().any(|b| b == &Some(true)) { continue }
                for b in &mut buf[window.clone()][..offset] {
                    *b = false;
                }

                if self.bits[offset..][..h].iter().any(|b| b == &Some(false)) { continue }
                for b in &mut buf[window.clone()][offset..][..h] {
                    *b = true;
                }

                if rest.len() == 0 {
                    RowView { pattern: rest, bits: &self.bits[offset+h..], freedom: self.freedom - offset}
                        .enumerate(buf, window.start + offset + h .., f);
                } else {
                    if self.bits[offset + h] == Some(true) { continue; }
                    buf[window.clone()][offset + h] = false;
                    RowView { pattern: rest, bits: &self.bits[offset+h+1..], freedom: self.freedom - offset }
                        .enumerate(buf, window.start + offset + h + 1 .., f);
                }

            }
        } else {
            for b in buf[window.clone()].iter_mut() { *b = false }
            if self.compatible(&buf[window]) {
                f(&buf)
            }
        }

    }

}

/*
fn pretty(row: &[bool]) -> String {
    row.iter().map(|&f| if f {'#'} else {'.'}).collect()
}
*/

#[test]
fn test_optimization() {
    let row: Row = "#??##????#??? 8,1,1".parse().unwrap();
    let row2 = row.clone();
    assert_eq!(row.count(), row2.count());
}
fn main() -> Result<()> {

    let rows: Vec<Row> = stdin().lines().map(|l| {
        l.unwrap().parse().unwrap()
    }).collect();

    {
        let mut rows = rows.clone();

        for row in &mut rows {
            row.optimize();
        }

        let total: usize = rows.iter().map(Row::count).sum();
        println!("Part 1: Total: {total}");

    }
    

    let mut rows: Vec<Row> = rows.into_iter().map(|r| r.multiply(5)).collect();
    for r in &mut rows { r.optimize() }

    let total: usize = rows.iter().map(Row::count).sum();
    println!("Part 2: Total: {total}");

    Ok(())

}
